from dataclasses import asdict
from datetime import datetime
from typing import Any, Tuple

from ._data import (
    OBS_TIME_T,
    Observatory,
    TopoCentricSunPositionResult,
    safely_from_dict,
)
from ._fseg import SPA_Analyzer


class SunEarthAnalyzer(object):
    """
    Interface class for sun-earth-analysis

    Args:
        algorithm (str): currently supports 'SPA'; Defaults to 'SPA'
    """

    def __init__(self, algorithm: str = "SPA") -> None:
        self._algorithm = algorithm
        self._obs = Observatory(0.0, 0.0, 0.0, 0.0)
        self._load_algorithm()

    def _load_algorithm(self):
        if self._algorithm.upper() == "SPA":
            self._impl = SPA_Analyzer()
        else:
            raise ValueError(f"Unknown algorithm {self._algorithm}")

    def set_observatory(self, **kwargs) -> None:
        obs_keys = vars(self._obs)
        for k, v in kwargs.items():
            if k in obs_keys:
                setattr(self._obs, k, v)

        self._impl.set_observatory(**asdict(self._obs))

    def get_observatory(self) -> Observatory:
        if self.has_set_observatory():
            return safely_from_dict(self._impl.get_observatory(), Observatory)
        else:
            return self._obs

    def has_set_observatory(self) -> bool:
        """
        check whether observatory is set for analyzer

        Returns:
            bool: True for set

        Examples:
            >>> sea = SunEarthAnalyzer()
            >>> sea.has_set_observatory()
            False
            >>> sea.sun_position_at(2020,5,13,17,15,30)
            Traceback (most recent call last):
              ...
            RuntimeError: Observatory has not set

            >>> d = {'timezone': -7.0, 'longitude': -105.1786, 'latitude': 39.742476,
            ...      'elevation': 1830.14, 'foo': 100.0}
            >>> sea.set_observatory(**d)
            >>> sea.has_set_observatory()
            True
        """
        return self._impl.has_set_observatory()

    def sun_position_at(
        self, dt: Any = None, DEBUG: bool = False, **kwargs
    ) -> Tuple[OBS_TIME_T, TopoCentricSunPositionResult]:
        """
        API for calculating sun position at a time

        Args:
            dt (Any, optional): supports str and datetime format of observation
                time. Defaults to None, when kwargs of (year, month, day, hour,
                minute, second) must be given.
            DEBUG (bool, optional): when set, print result in multiline format.
                Defaults to False.

        Raises:
            ValueError: invalid inputs for observation time or
                failed at validation stage of the implemented algorithm

        Returns:
            (dt,sp) (Tuple[OBS_TIME_T, TopoCentricSunPositionResult]):
                dt (year,month,day,hour,minute,second): parsed observation time
                sp: topocentric solar position with julian day

        Examples:
            >>> sea = SunEarthAnalyzer()
            >>> sea.set_observatory(
            ...     longitude=-105.1786, latitude=39.742476, elevation=1830.14,
            ...     timezone=-7.0, delta_ut1=0, delta_t=67,
            ...     pressure=820, temperature=11, atmos_refract=0.5667,
            ... )
            >>> dt, sp = sea.sun_position_at("2003-10-17 12:30:30")
            >>> dt
            (2003, 10, 17, 12, 30, 30)
            >>> sp
            TopoCentricSunPositionResult(zenith=50.11162202402972,
            ... azimuth=194.34024051019162, julian_day=2452930.312847222)
        """
        try:
            if dt is None:
                _year, _month = int(kwargs["year"]), int(kwargs["month"])
                _day, _hour = int(kwargs["day"]), int(kwargs["hour"])
                _minute, _second = int(kwargs["minute"]), int(kwargs["second"])
            else:
                if isinstance(dt, str):
                    dt = datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")

                if isinstance(dt, datetime):
                    _year, _month, _day = dt.year, dt.month, dt.day
                    _hour, _minute, _second = dt.hour, dt.minute, dt.second
                else:
                    raise TypeError(f"Invalid type for dt: {type(dt)}")
        except Exception:
            raise ValueError(f"Invalid argument: dt={dt}, {kwargs}")

        sp = self._sun_position_at(_year, _month, _day, _hour, _minute, _second)

        if DEBUG:
            obs = self.get_observatory()
            print("----------INPUT----------")
            print("Year:           %d" % _year)
            print("Month:          %d" % _month)
            print("Day:            %d" % _day)
            print("Hour:           %d" % _hour)
            print("Minute:         %d" % _minute)
            print("Second:         %d" % _second)
            print("Timezone:       %.6f" % obs.timezone)
            print("Longitude:      %.6f" % obs.longitude)
            print("Latitude:       %.6f" % obs.latitude)
            print("Elevation:      %.6f" % obs.elevation)
            print("Pressure:       %.6f" % obs.pressure)
            print("Temperature:    %.6f" % obs.temperature)
            print("Atmos_Refract:  %.6f" % obs.atmos_refract)
            print("Delta T:        %.6f" % obs.delta_t)
            print("----------OUTPUT----------")
            if sp.julian_day:
                print("Julian Day:    %.6f" % sp.julian_day)
            print("Zenith:        %.6f degrees" % sp.zenith)
            print("Azimuth:       %.6f degrees" % sp.azimuth)

        dt = (_year, _month, _day, _hour, _minute, _second)
        return dt, sp

    def _sun_position_at(
        self, year: int, month: int, day: int, hour: int, minute: int, second: int
    ) -> TopoCentricSunPositionResult:
        """
        directly call c++ library, when performance is critical
        """
        zenith, azimuth, julian_day = self._impl.calc_sun_position_at(
            year, month, day, hour, minute, second
        )

        return TopoCentricSunPositionResult(zenith, azimuth, julian_day)

    def __repr__(self) -> str:
        return f"SunEarthAnalyzer(algorithm={self._algorithm})"



from ._version import __version__  # isort: skip
from ._analyzer import SunEarthAnalyzer
from ._data import Observatory, TopoCentricSunPositionResult

__all__ = [
    "__version__",
    "SunEarthAnalyzer",
    "Observatory",
    "TopoCentricSunPositionResult",
]



from dataclasses import dataclass, fields, is_dataclass
from typing import Dict, Optional, Tuple, Type, TypeAlias, TypeVar


@dataclass
class Observatory:
    longitude: float
    latitude: float
    elevation: float
    timezone: float
    delta_ut1: float = 0
    delta_t: float = 0
    pressure: float = 0
    temperature: float = 0
    atmos_refract: float = 0


@dataclass
class TopoCentricSunPositionResult:
    zenith: float
    azimuth: float
    julian_day: Optional[float] = None


@dataclass
class TopoCentricSunRiseTransitSetResult:
    suntransit: float
    sunrise: float
    sunset: float


# @dataclass
# class SunPathResult:
#     pass

OBS_TIME_T: TypeAlias = Tuple[int, int, int, int, int, int]

_Data = TypeVar(
    "_Data",
    Observatory,
    TopoCentricSunPositionResult,
    TopoCentricSunRiseTransitSetResult,
)


def safely_from_dict(d: Dict, cls: Type[_Data]) -> _Data:
    """
    generate dataclass from dict by dropping irrelevant keys

    Args:
        d (Dict): data dict
        cls (Type[_Data]): data class type

    Raises:
        TypeError: failed to generate data class

    Returns:
        _Data: data class of type <cls>

    Examples:
        >>> d = {'timezone': -7.0, 'longitude': -105.1786, 'latitude': 39.742476,
        ... 'elevation': 1830.14, 'foo': 100.0}
        >>> safely_from_dict(d, Observatory)
        Observatory(longitude=-105.1786, latitude=39.742476, elevation=1830.14,
        timezone=-7.0, delta_ut1=0, delta_t=0, pressure=0, temperature=0,
        atmos_refract=0)

    """
    if is_dataclass(cls):
        field_names = {f.name for f in fields(cls)}
        filtered_d = {k: v for k, v in d.items() if k in field_names}
        return cls(**filtered_d)
    else:
        raise TypeError(f"Invalid class type: {cls}")



import os
import re
import subprocess
import sys
import sysconfig

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection & inclusion of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # SDIST_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        # In this example, we pass in the version to C++. You might not need to.
        cmake_args += [f"-DSDIST_VERSION_INFO={self.distribution.get_version()}"]

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator or cmake_generator == "Ninja":
                try:
                    import ninja  # noqa: F401

                    ninja_executable_path = os.path.join(ninja.BIN_DIR, "ninja")
                    cmake_args += [
                        "-GNinja",
                        f"-DCMAKE_MAKE_PROGRAM:FILEPATH={ninja_executable_path}",
                    ]
                except ImportError:
                    pass

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}",
                    # DLL are treated as RUNTIME in cmake
                    f"-DCMAKE_RUNTIME_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}",
                    # .exp and .lib files to be packaged into wheel too
                    f"-DCMAKE_ARCHIVE_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}",
                ]
                build_args += ["--config", cfg]

            assert cfg == "Release", "MSVC only support Release build-type"
            sysconfig.get_config_vars()["Py_DEBUG"] = False

        if sys.platform.startswith("darwin"):
            macosx_version_min = sysconfig.get_config_var("MACOSX_DEPLOYMENT_TARGET")
            if macosx_version_min:
                cmake_args += [
                    "-DCMAKE_OSX_DEPLOYMENT_TARGET={}".format(macosx_version_min)
                ]

            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += [f"-j{self.parallel}"]

        build_temp = os.path.join(self.build_temp, ext.name)
        if not os.path.exists(build_temp):
            os.makedirs(build_temp)

        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, cwd=build_temp)
        subprocess.check_call(["cmake", "--build", "."] + build_args, cwd=build_temp)


# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.
setup(
    packages=find_packages(where="src/python"),
    package_dir={"fseg": "src/python/fseg"},
    ext_package="fseg",
    ext_modules=[CMakeExtension("_fseg")],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    python_requires=">=3.10",
    platforms=[
        "cp310-macosx-10_9-x86_64",
        "cp310-manylinux2014_x86_64",
        "cp310-win_amd64",
    ],
)
